<?php
/**
 * This file is part of the SetaPDF-Core Component
 *
 * @copyright  Copyright (c) 2017 Setasign - Jan Slabon (https://www.setasign.com)
 * @category   SetaPDF
 * @package    SetaPDF_Core
 * @subpackage Document
 * @license    https://www.setasign.com/ Commercial
 * @version    $Id: Document.php 1067 2017-06-26 15:14:38Z jan.slabon $
 */

/**
 * A class representing a PDF document
 *
 * This class represents a PDF document in all SetaPDF components.
 * It offers the main functionalities for managing objects, cross
 * reference tables and writers of the document instance.
 *
 * It also tracks changes of objects and security handlers.
 *
 * @copyright  Copyright (c) 2017 Setasign - Jan Slabon (https://www.setasign.com)
 * @category   SetaPDF
 * @package    SetaPDF_Core
 * @subpackage Document
 * @license    https://www.setasign.com/ Commercial
 */
class SetaPDF_Core_Document
    implements SplObserver, SetaPDF_Core_WriteInterface, SetaPDF_Core_Type_Owner
{
    /**
     * State constant
     *
     * @var string
     */
    const STATE_NONE = 'none';

    /**
     * State constant
     *
     * @var string
     */
    const STATE_WRITING_BODY = 'writingBody';

    /**
     * State constant
     *
     * @var string
     */
    const STATE_WRITING_XREF = 'writingXRef';

    /**
     * State constant
     *
     * @var string
     */
    const STATE_SAVED = 'saved';

    /**
     * State constant
     *
     * @var string
     */
    const STATE_FINISHED = 'finished';

    /**
     * State constant
     *
     * @var string
     */
    const STATE_CLEANED_UP = 'cleanedUp';

    /**
     * Save method constant defining an incremental update
     *
     * @var string
     */
    const SAVE_METHOD_UPDATE = 1;

    /**
     * Save method constant defining a rewrite by resolving objects starting at the root object
     *
     * @var string
     */
    const SAVE_METHOD_REWRITE = 0;

    /**
     * Save method constant defining a rewrite by writing all available objects
     *
     * @var string
     */
    const SAVE_METHOD_REWRITE_ALL = -1;

    /**
     * A counter for generating unique instance identifications
     *
     * @var integer
     */
    static protected $_instanceCounter = 0;

    /**
     * A random prefix for generating unique instance identifications
     *
     * @var string
     */
    static protected $_instanceIdentPrefix = null;

    /**
     * Incremental update or rewrite the document
     *
     * @see SetaPDF_Core_Document::save()
     * @var integer
     */
    protected $_saveMethod = self::SAVE_METHOD_REWRITE;

    /**
     * A flag defining the state of the document object instance
     *
     * @var string
     */
    protected $_state = self::STATE_NONE;

    /**
     * PDF version
     *
     * @see SetaPDF_Core_Document::setPdfVersion()
     * @var string
     */
    protected $_pdfVersion = '1.3';

    /**
     * An instance of a cross reference
     *
     * If the document is created of an existing one
     * this will be an instance of SetaPDF_Core_Parser_CrossReferenceTable
     *
     * @var SetaPDF_Core_Document_CrossReferenceTable
     */
    protected $_xref;

    /**
     * Defines if the cross reference table will be compressed
     *
     * @see SetaPDF_Core_Document::setCompressXref()
     * @var boolean
     */
    protected $_compressXref = false;

    /**
     * Current/max object id
     *
     * @var integer
     */
    protected $_maxObjId = 0;

    /**
     * Newly created or resolved objects
     *
     * @var array
     */
    protected $_objects = array();

    /**
     * Array for information about object streams
     *
     * @var array
     */
    protected $_objectStreams = array();

    /**
     * The parser object used for parsing object streams
     *
     * @var SetaPDF_Core_Parser_Pdf
     */
    protected $_objectStreamsParser;

    /**
     * The trailer dictionary
     *
     * @var SetaPDF_Core_Type_Dictionary
     */
    protected $_trailer;

    /**
     * Flag defining if the trailer was touched/changed
     *
     * @var boolean
     */
    protected $_trailerChanged = false;

    /**
     * Changed objects
     *
     * @var array
     */
    protected $_changedObjects;

    /**
     * Referenced objects
     *
     * This array holds information about objects to which
     * references were written. Needed to create deep
     * copies of an object from one to another document
     *
     * @var array
     */
    protected $_referencedObjects = array();

    /**
     * Blocked referenced objects
     *
     * This array holds objects which should NOT be automatically
     * resolved.
     *
     * @var array
     */
    protected $_blockedReferencedObjects = array();

    /**
     * A relation between objects and ids
     *
     * @var array
     */
    protected $_objectsToIds = array();

    /**
     * The writer instance
     *
     * @see SetaPDF_Core_Document::setWriter()
     * @see SetaPDF_Core_Document::__construct()
     * @var SetaPDF_Core_Writer_WriterInterface
     */
    protected $_writer = null;

    /**
     * The parser object of the existing document
     *
     * @var SetaPDF_Core_Parser_Pdf
     */
    protected $_parser;

    /**
     * The indirect object which is currently written
     *
     * @var SetaPDF_Core_Type_IndirectObject
     */
    protected $_currentObject;

    /**
     * The object id and generation number of the currently written object
     *
     * @var array
     */
    protected $_currentObjectData;

    /**
     * The security handler of the existing document
     *
     * @var SetaPDF_Core_SecHandler_SecHandlerInterface
     */
    protected $_secHandlerIn;

    /**
     * The security handler of the new document
     *
     * @var SetaPDF_Core_SecHandler_SecHandlerInterface
     */
    protected $_secHandler;

    /**
     * Identification of a document instance
     *
     * @var string
     */
    protected $_instanceIdent;

    /**
     * Documents catalog instance
     *
     * @var SetaPDF_Core_Document_Catalog
     */
    protected $_catalog;

    /**
     * The documents info object instance
     *
     * @var SetaPDF_Core_Document_Info
     */
    protected $_info;

    /**
     * Flag saying that objects should be cleaned up automatically
     *
     * @var boolean
     */
    protected $_cleanUpObjects = true;

    /**
     * A method/function which should be called to fill the document body
     *
     * @var callback
     */
    protected $_fileBodyMethod = null;

    /**
     * Defining whether the PDF objects should be written at once or object by object
     *
     * @var boolean
     */
    protected $_directWrite = false;

    /**
     * Flag saying that write callbacks are in use
     *
     * @var boolean
     */
    protected $_useWriteCallbacks = false;

    /**
     * Array of write callbacks
     *
     * @var array
     */
    protected $_writeCallbacks = array();

    /**
     * The none permanent file identifier
     *
     * @var string
     */
    protected $_newFileIdentifier = null;

    /**
     * Defines if referenced objects should be cached or not
     *
     * @var boolean
     */
    protected $_cacheReferencedObjects = false;

    /**
     * An array of callbacks that should be called before the save method is executed.
     *
     * @var array
     */
    protected $_beforeSaveCallbacks = array();



    /**
     * The constructor.
     *
     * @param SetaPDF_Core_Writer_WriterInterface $writer The writer to which the document should be written
     */
    public function __construct(SetaPDF_Core_Writer_WriterInterface $writer = null)
    {
        if (null === self::$_instanceIdentPrefix) {
            self::$_instanceIdentPrefix = md5(microtime() . mt_rand());
        }

        $this->setWriter($writer);
        $this->_xref = new SetaPDF_Core_Document_CrossReferenceTable();
        $this->_changedObjects = array();
        $this->_trailer = new SetaPDF_Core_Type_Dictionary();
        $this->_trailer->offsetSet(null, new SetaPDF_Core_Type_Dictionary_Entry(
            new SetaPDF_Core_Type_Name('Size', true),
            new SetaPDF_Core_Type_Numeric($this->_xref->getSize())
        ));

        $this->_instanceIdent = self::$_instanceIdentPrefix
            . '-'
            . self::$_instanceCounter++;
    }

    /**
     * Set the writer object.
     *
     * A writer instance can only be set prior the first call to {@link SetaPDF_Core_Document::save() save()} or
     * after a {@link SetaPDF_Core_Document::finish() finish()} call.
     *
     * @param SetaPDF_Core_Writer_WriterInterface $writer The new writer object
     * @throws BadMethodCallException
     * @return SetaPDF_Core_Document
     */
    public function setWriter(SetaPDF_Core_Writer_WriterInterface $writer = null)
    {
        if (
            $this->_state !== SetaPDF_Core_Document::STATE_FINISHED &&
            $this->_writer instanceof SetaPDF_Core_Writer_WriterInterface &&
            $this->_writer->getStatus() != SetaPDF_Core_Writer::INACTIVE
        ) {
            throw new BadMethodCallException(
                'It is not possible to change the writer after starting output.'
            );
        }

        $this->_writer = $writer;

        return $this;
    }


    /**
     * Get the parser object.
     *
     * @return SetaPDF_Core_Parser_Pdf|null
     */
    public function getParser()
    {
        return $this->_parser;
    }

    /**
     * Return the current object state.
     *
     * @return string
     */
    public function getState()
    {
        return $this->_state;
    }


    /**
     * Returns the PDF version of the document.
     *
     * @return string
     */
    public function getPdfVersion()
    {
        return $this->_pdfVersion;
    }

    /**
     * Set the PDF version of the document.
     *
     * @param string $pdfVersion The pdf version
     * @return void
     */
    public function setPdfVersion($pdfVersion)
    {
        $this->_pdfVersion = is_float($pdfVersion) ? sprintf('%.1F', $pdfVersion) : $pdfVersion;

        if ($this->_saveMethod === self::SAVE_METHOD_UPDATE) {
            $this->getCatalog()->setVersion($this->_pdfVersion);
        }
    }

    /**
     * Set the minimal PDF version.
     *
     * @param string $minPdfVersion The minimal pdf version
     */
    public function setMinPdfVersion($minPdfVersion)
    {
        if (version_compare($this->_pdfVersion, $minPdfVersion, '<')) {
            $this->setPdfVersion($minPdfVersion);
        }
    }

    /**
     * Get the catalog object.
     *
     * @return SetaPDF_Core_Document_Catalog
     */
    public function getCatalog()
    {
        if (null === $this->_catalog)
            $this->_catalog = new SetaPDF_Core_Document_Catalog($this);

        return $this->_catalog;
    }

    /**
     * Get the documents info object.
     *
     * @return SetaPDF_Core_Document_Info
     */
    public function getInfo()
    {
        if (null === $this->_info)
            $this->_info = new SetaPDF_Core_Document_Info($this);

        return $this->_info;
    }

    /**
     * Implementation of the observer pattern.
     *
     * This method is automatically called if an observed object
     * was changed.
     *
     * @param SplSubject $subject The SplSubject notifying the observer of an update.
     * @return void
     */
    public function update(SplSubject $subject)
    {
        if ($subject instanceof SetaPDF_Core_Type_IndirectObject) {
            $objectData = $this->getIdForObject($subject);
            if (!isset($this->_changedObjects[$objectData[0]])) {
                $this->_changedObjects[$objectData[0]] = $subject;
            }

        } elseif ($subject === $this->_trailer) {
            $this->_trailerChanged = true;
        }
    }

    /**
     * Define whether the cross reference should be compressed or not.
     *
     * By default the SetaPDF-Core component writes the cross-reference in the standard format or in the format which
     * is defined in the source document, if any available.
     *
     * @param boolean $compressXref Pass true to enforce that the cross reference will be compressed. Pass false to
     *                              enforce a standard uncompressed cross reference table.
     * @return void
     * @throws BadMethodCallException
     */
    public function setCompressXref($compressXref)
    {
        if ($this->_state !== self::STATE_NONE) {
            throw new BadMethodCallException(
                'It is not possible to change the compression of the xref table after an initial save call.'
            );
        }

        $this->_compressXref = (boolean)$compressXref;
        if (true === $this->_compressXref) {
            $this->setMinPdfVersion('1.5');
        }
    }

    /**
     * Get the cross reference object.
     *
     * @return SetaPDF_Core_Document_CrossReferenceTable
     */
    public function getXref()
    {
        return $this->_xref;
    }

    /**
     * Set the behavior if the cleanUp()-methods of objects get called automatically.
     *
     * @param boolean $cleanUpObjects The flag status
     */
    public function setCleanUpObjects($cleanUpObjects)
    {
        $this->_cleanUpObjects = (boolean)$cleanUpObjects;
    }

    /**
     * Define if referenced objects should be cached or not.
     *
     * @param boolean $cacheReferencedObjects The flag status
     */
    public function setCacheReferencedObjects($cacheReferencedObjects)
    {
        $this->_cacheReferencedObjects = (boolean)$cacheReferencedObjects;
    }

    /**
     * Says that referenced objects get cached or not.
     *
     * @return boolean
     */
    public function getCacheReferencedObjects()
    {
        return $this->_cacheReferencedObjects;
    }

    /**
     * Checks whether a security handler is attached to this document.
     *
     * @return boolean
     */
    public function hasSecHandler()
    {
        return null !== $this->_secHandler;
    }

    /**
     * Alias for hasSecHandler().
     *
     * @deprecated
     * @return bool
     */
    public function hasSecurityHandler()
    {
        return $this->hasSecHandler();
    }

    /**
     * Returns the security handler of the original document.
     *
     * @return null|SetaPDF_Core_SecHandler_SecHandlerInterface
     */
    public function getSecHandlerIn()
    {
        return $this->_secHandlerIn;
    }

    /**
     * Get the security handler for the output document.
     *
     * @return SetaPDF_Core_SecHandler_SecHandlerInterface|SetaPDF_Core_SecHandler_Standard
     */
    public function getSecHandler()
    {
        return $this->_secHandler;
    }

    /**
     * Defines whether the PDF objects should be written individually (true) or after assembling a single string (false).
     *
     * @param bool $directWrite
     */
    public function setDirectWrite($directWrite)
    {
        $this->_directWrite = (boolean)$directWrite;
    }

    /**
     * Gets whether the PDF objects should be written individually (true) or after assembling a single string (false).
     *
     * @return bool
     */
    public function getDirectWrite()
    {
        return $this->_directWrite;
    }

    /**
     * Writes content to the attached writer.
     *
     * @param string $s
     * @return mixed
     * @throws SetaPDF_Core_Exception
     */
    public function write($s)
    {
        if (null !== $this->_writer)
            return $this->_writer->write($s);

        throw new SetaPDF_Core_Exception(
            'No writer object known!'
        );
    }

    /**
     * Implementation of the SetaPDF_Core_Type_Owner interface.
     *
     * @return $this
     */
    public function getOwnerPdfDocument()
    {
        return $this;
    }
}