<?php
/**
 * This file is part of the SetaPDF-Core Component
 *
 * @copyright  Copyright (c) 2017 Setasign - Jan Slabon (https://www.setasign.com)
 * @category   SetaPDF
 * @package    SetaPDF_Core
 * @subpackage SecHandler
 * @license    https://www.setasign.com/ Commercial
 * @version    $Id: SecHandler.php 1052 2017-05-18 09:49:46Z jan.slabon $
 */

/**
 * Main class for PDF security handlers
 *
 * @copyright  Copyright (c) 2017 Setasign - Jan Slabon (https://www.setasign.com)
 * @category   SetaPDF
 * @package    SetaPDF_Core
 * @subpackage SecHandler
 * @license    https://www.setasign.com/ Commercial
 */
abstract class SetaPDF_Core_SecHandler
{
    /**
     * Standard Security Handler
     *
     * @var string
     */
    const STANDARD = 'standard';

    /**
     * Public Key Security Handler
     *
     * @var string
     */
    const PUB_KEY = 'publicKey';

    /**
     * Encryption constant
     *
     * @var string
     */
    const ARCFOUR = 4;

    /**
     * Encryption constant
     *
     * @var string
     */
    const ARCFOUR_40 = 12; // 8 | 4

    /**
     * Encryption constant
     *
     * @var string
     */
    const ARCFOUR_128 = 20; // 16 | 4

    /**
     * Encryption constant
     *
     * @var string
     */
    const AES = 32;

    /**
     * Encryption constant
     *
     * @var string
     */
    const AES_128 = 96; // 64 | 32

    /**
     * Encryption constant
     *
     * @var string
     */
    const AES_256 = 160; // 128 | 32

    /**
     * Permission constant.
     *
     * For handlers of revision 2: Print the document.
     *
     * Handlers of a revision of 3 or greater: Print the document (possibly not at the highest quality level, depending
     * on whether {@link SetaPDF_Core_SecHandler::PERM_DIGITAL_PRINT} is also set).
     *
     * @see PDF 32000-1:2008 - Table 22 - User access permissions
     * @var integer
     */
    const PERM_PRINT = 4; // 3

    /**
     * Permission constant.
     *
     * Modify the contents of the document by operations other than those controlled by
     * {@link SetaPDF_Core_SecHandler::PERM_ANNOT}, {@link SetaPDF_Core_SecHandler::PERM_FILL_FORM} and
     * {@link SetaPDF_Core_SecHandler::PERM_ASSEMBLE}.
     *
     * @see PDF 32000-1:2008 - Table 22 - User access permissions
     * @var integer
     */
    const PERM_MODIFY = 8; // 4

    /**
     * Permission constant.
     *
     * For handlers of revision 2: Copy or otherwise extract text and graphics from the document, including extracting
     * text and graphics (in support of accessibility to users with disabilities or for other purposes).
     *
     * For handlers of revision 3 or greater: Copy or otherwise extract text and graphics from the document by
     * operations other than that controlled by bit {@link SetaPDF_Core_SecHandler::PERM_ACCESSIBILITY}.
     *
     * @see PDF 32000-1:2008 - Table 22 - User access permissions
     * @var integer
     */
    const PERM_COPY = 16; // 5

    /**
     * Permission constant.
     *
     * Add or modify text annotations, fill in interactive form fields, and, if {@link SetaPDF_Core_SecHandler::PERM_MODIFY}
     * is also set, create or modify interactive form fields (including signature fields).
     *
     * @see PDF 32000-1:2008 - Table 22 - User access permissions
     * @var integer
     */
    const PERM_ANNOT = 32; // 6

    /**
     * Permission constant.
     *
     * For handlers of revision 3 or greater: Fill in existing interactive form fields (including signature fields),
     * even if {@link SetaPDF_Core_SecHandler::PERM_ANNOT} is not set.
     *
     * @see PDF 32000-1:2008 - Table 22 - User access permissions
     * @var integer
     */
    const PERM_FILL_FORM = 256; // 9

    /**
     * Permission constant.
     *
     * For handlers of revision 3 or greater: Extract text and graphics (in support of accessibility to users with
     * disabilities or for other purposes).
     *
     * @see PDF 32000-1:2008 - Table 22 - User access permissions
     * @var integer
     */
    const PERM_ACCESSIBILITY = 512; // 10

    /**
     * Permission constant.
     *
     * For handlers of revision 3 or greater: Assemble the document (insert, rotate, or delete pages and create
     * bookmarks or thumbnail images), even if {@link SetaPDF_Core_SecHandler::PERM_MODIFY} is not set.
     *
     * @see PDF 32000-1:2008 - Table 22 - User access permissions
     * @var integer
     */
    const PERM_ASSEMBLE = 1024; // 11

    /**
     * Permission constant.
     *
     * Print the document to a representation from which a faithful digital copy of the PDF content could be generated.
     * When this is not set (and {@link SetaPDF_Core_SecHandler::PERM_PRINT} is set), printing is limited to a low-level
     * representation of the appearance, possibly of degraded quality.
     *
     * @see PDF 32000-1:2008 - Table 22 - User access permissions
     * @var integer
     */
    const PERM_DIGITAL_PRINT = 2048; // 12

    /**
     * User auth mode
     *
     * @var string
     */
    const USER = 'user';

    /**
     * Owner auth mode
     *
     * @var string
     */
    const OWNER = 'owner';

    /**
     * The encryption engine to use (mcrypt or openssl).
     *
     * @var string
     */
    static public $engine = 'openssl';

    /**
     * Checks a permission against the security handler of a document.
     *
     * @param SetaPDF_Core_Document $document The document instance
     * @param integer $permission Permission to check
     * @param null|string $message Custom error message
     * @return bool
     * @throws SetaPDF_Core_SecHandler_Exception if no rights are granted for the permission.
     */
    static public function checkPermission(SetaPDF_Core_Document $document, $permission, $message = null)
    {
        return true;
    }

}